/* fumied.c */

#include <signal.h>
#include <pwd.h>
#include "fumied.h"

static void usage();

char fumied_pid_path[128];
char etc_hosts[128];
char temp_hosts[128];

TIME cur_time;

u_int local_port;
u_int remote_port;

int log_priority;
#ifdef DEBUG
int log_perror = -1;
#else
int log_perror = 1;
#endif

#define DEFAULT_IF	"eth0"

int main(int argc, char **argv)
{
  int i, status;
  struct servent *ent;
  int  daemonFlag = 1;		/* if 1: work as a client only */
  int  killFlag = 0;		/* if 1: kill the running proc and exit */
  int  clientOnly = 0;		/* if 1: work as a client only */
  int  serverOnly = 0;		/* if 1: work as a server only */
#ifndef DEBUG
  int pidfilewritten = 0;
  int pid;
  char pbuf [20];
#endif

  /* Initially, log errors to stderr as well as to syslogd. */
  openlog ("fumied", LOG_NDELAY, FUMIED_LOG_FACILITY);

  srand((u_int)time(NULL));
  umask(0);					/* clear umask */

  while ( *++argv )
    {
      if ( **argv == '-' )
	{
	  switch ( argv[0][1] )
	    {
	    case 'd':
#ifndef DEBUG
	      daemonFlag = 0;
#endif
	      break;
	    case 'k':
	      killFlag = 1;	/* kill running process and exit */
	      break;
	    case 's':
	      serverOnly = 1;	/* if 1: work as a server only */
	      break;
	    case 'c':
	      clientOnly = 1;	/* if 1: work as a client only */
	      break;
	    default:
	      usage();
	    }
	}
      else
	{
	  break;
	}
    }

#ifndef DEBUG
  if ( getuid() != 0 && geteuid() != 0 )
    {
      error ("Must be root.");
    }
  else
    {
      strcpy(fumied_pid_path, _SUID_PATH_FUMIED_PID);
      strcpy(etc_hosts, _SUID_ETC_HOSTS);
      strcpy(temp_hosts, _SUID_TEMP_HOSTS);
      mktemp(temp_hosts);
    }

  if ( killFlag )
    {
      /* Read previous pid file. */
      if ((i = open (fumied_pid_path, O_RDONLY)) >= 0)
	{
	  status = read (i, pbuf, (sizeof pbuf) - 1);
	  close (i);
	  pbuf [status] = 0;
	  pid = atoi (pbuf);
	
	  /* If the previous server process is not still running,
	     write a new pid file immediately. */
	  if (pid && kill (pid, SIGTERM) == 0)
	    {
	      if (unlink (fumied_pid_path) == 0)
		{
		  note ("Previous process %d is killed.", pid);
		  return 0;
		}
	      error ("Previous process isn't found.");
	    }
	}
      else
	{
	  error ("fumied isn't running or previous process isn't found.");
	}
    }
#endif

  /* Read previous pid file. */
  if ((i = open (fumied_pid_path, O_RDONLY)) >= 0)
    {
      status = read (i, pbuf, (sizeof pbuf) - 1);
      close (i);
      pbuf [status] = 0;
      pid = atoi (pbuf);

      /* If the previous server process is not still running,
	 write a new pid file immediately. */
      if (pid != 0)
	{
	  error ("fumied is already running as pid %d.", pid);
	}
    }

  if ( !serverOnly )
    {
      getIfInfo(DEFAULT_IF);
      fumiedMsgInit(DEFAULT_IF);
      fumiedClient();
    }

  if (clientOnly)
    {
      return 0;
    }

  /* Default to the FUMIED port. */
  if (!local_port)
    {
      ent = getservbyname ("fumie", "udp");
      if (!ent)
	local_port = htons (FUMIED_SERVER_PORT);
      else
	local_port = ent -> s_port;
      endservent ();
    }
  
  remote_port = htons (ntohs (local_port) + 1);

  /* Get the current time... */
  GET_TIME (&cur_time);

  /* Discover all the network interfaces and initialize them. */
  discover_interfaces (DISCOVER_SERVER);

#ifndef DEBUG
  if (daemonFlag)
    {
      /* First part of becoming a daemon... */
      if ((pid = fork ()) < 0)
	error ("Daemon couldn't fork: %m");
      else if (pid)
	exit (0);
    }

  /* Read previous pid file. */
  if ((i = open (fumied_pid_path, O_RDONLY)) >= 0)
    {
      status = read (i, pbuf, (sizeof pbuf) - 1);
      close (i);
      pbuf [status] = 0;
      pid = atoi (pbuf);

      /* If the previous server process is not still running,
	 write a new pid file immediately. */
      if (pid && kill (pid, 0) < 0)
	{
	  unlink (fumied_pid_path);
	  if ((i = open (fumied_pid_path, O_WRONLY | O_CREAT, 0640)) >= 0)
	    {
	      sprintf (pbuf, "%d\n", (int)getpid ());
	      write (i, pbuf, strlen (pbuf));
	      close (i);
	      pidfilewritten = 1;
	    }
	}
    }

  /* If we were requested to log to stdout on the command line,
     keep doing so; otherwise, stop. */
  if (log_perror == -1)
    log_perror = 1;
  else
    log_perror = 0;

  if (daemonFlag)
    {
      /* Become session leader and get pid... */
      close (0);
      close (1);
      close (2);
      pid = setsid ();
    }

  if (!pidfilewritten)
    {
      unlink (fumied_pid_path);
      if ((i = open (fumied_pid_path, O_WRONLY | O_CREAT, 0640)) >= 0)
	{
	  sprintf (pbuf, "%d\n", (int)getpid ());
	  write (i, pbuf, strlen (pbuf));
	  close (i);
	  pidfilewritten = 1;
	}
    }
#endif /* !DEBUG */

  /* Set up the fumied packet handler... */
  fumied_packet_handler = do_packet;

  /* Receive packets and dispatch them... */
  dispatch ();

  /* Not reached */
  return 0;
}

/* Print usage message. */

static void usage()
{
  error ("Usage: fumied [-k] [-c] [-s] [-d]");
}

void cleanup()
{
}
