/* socket-if.c */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <arpa/inet.h>
#include <net/route.h>
#include <errno.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include "fumied.h"

struct ifinfo Ifbuf;

void
getIfInfo(char *ifname)
{
  int s;
  struct ifreq ifr;
  struct ifinfo *ifinfo;

  ifinfo = &Ifbuf;
  bzero((char *)ifinfo, sizeof(Ifbuf));
  strncpy(ifinfo->ifname, ifname, sizeof(Ifbuf.ifname));

  strcpy(ifr.ifr_name, ifinfo->ifname);
  if ( (s = socket(AF_INET, SOCK_DGRAM, 0)) < 0 )
    {
      error("socket (getIfInfo)");
    }

  /* configure interface
   */
  if ( ioctl(s, SIOCGIFADDR, &ifr) < 0 )
    {
      error("ioctl SIOCGIFADDR (getIfInfo)");
    }
  ifinfo->addr = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr;
  if ( ioctl(s, SIOCGIFBRDADDR, &ifr) < 0 )
    {
      error("ioctl SIOCGIFBRDADDR (getIfInfo)");
    }
  ifinfo->bcast = ((struct sockaddr_in *)&ifr.ifr_broadaddr)->sin_addr.s_addr;
  if ( ioctl(s, SIOCGIFNETMASK, &ifr) < 0 )
    {
      error("ioctl SIOCGIFNETMASK (getIfInfo)");
    }
  ifinfo->mask = ((struct sockaddr_in *)&ifr.ifr_netmask)->sin_addr.s_addr;
  if ( ioctl(s, SIOCGIFFLAGS, &ifr) < 0 )
    {
      error("ioctl SIOCGIFFLAGS (getIfInfo)");
    }
  ifinfo->flags = ifr.ifr_flags;
  close(s);
}

void
setSockAddrIn(u_short port, u_long inaddr, struct sockaddr_in *saddr)
{
  bzero((char *)saddr, sizeof(*saddr));
  saddr->sin_family		= AF_INET;
  saddr->sin_addr.s_addr	= inaddr;
  saddr->sin_port		= port;
}

void
openSendSocket(struct sockaddr_in *addr, int *s)
{
  int optval = 1;

  if ( (*s = socket(AF_INET, SOCK_DGRAM, 0)) < 0 )
    {
      error("socket (openSendSocket)");
    }
  if ( setsockopt(*s, SOL_SOCKET, SO_BROADCAST, &optval, sizeof(optval)) < 0)
    {
      error("setsockopt (openSendSocket)");
    }
  /* use FUMIED client port for the source port because
   * some servers do NOT use the FUMIED client port
   * BUT use the UDP source  port number of the received
   * datagram for the destination UDP port when it responds
   * to clients.
   */
  addr->sin_port = htons(FUMIED_CLIENT_PORT);
#if 0
  addr->sin_port = htons(0);	/* appropriate unused port */
#endif
  if ( bind(*s, (struct sockaddr *)addr, sizeof(*addr))  < 0 )
    {
      error("bind (openSendSocket)");
    }
}

void
openRecvSocket(struct sockaddr_in *addr, int *s)
{
  int optval = 1;

  if ( (*s = socket(AF_INET, SOCK_DGRAM, 0)) < 0 )
    {
      error("socket (openRecvSocket)");
    }
  if ( setsockopt(*s, SOL_SOCKET, SO_BROADCAST, &optval, sizeof(optval)) < 0)
    {
      error("setsockopt (openRecvSocket)");
    }
  addr->sin_port = htons(FUMIED_CLIENT_PORT);
  if ( bind(*s, (struct sockaddr *)addr, sizeof(*addr))  < 0 )
    {
      error("bind (openRecvSocket)");
    }
}

int
rcvAndCheckFumiedMsg(int s, struct fumiedMessage *msg, u_long waitMsgType,
		    long timeout)
{
  fd_set fdset;
  time_t prevTime;
  struct sockaddr_in addr;
  struct timeval tm;
  int len;

  bzero((char *)msg, sizeof(*msg));
  bzero((char *)&addr, sizeof(addr));
  tm.tv_sec = 0;
  time(&prevTime); 
  while ( timeout > 0 )
    {
      FD_ZERO(&fdset);
      FD_SET(s, &fdset);
      tm.tv_usec  = timeout;
      if ( select(s+1, &fdset, (fd_set *)NULL, (fd_set *)NULL, &tm) < 0 )
	{
	  debug("select (rcvAndCheckFumiedMsg)");
	  return 0;			/* receive unsuccessful */
	}
      len = sizeof(*msg);
      if ( FD_ISSET(s, &fdset) )
	{
	  if ( recvfrom(s, (char *)msg, len, 0,
			(struct sockaddr *)&addr, &len) < 0 )
	    {
	      debug("recvfrom (rcvAndCheckFumiedMsg)");
	      return 0;			/* receive unsuccessful */
	    }
	  return 1;
	}
      timeout -= (time(NULL) - prevTime)*1000000;
      time(&prevTime);
    }
  return 0;				/* receive timeout */
}

int
waitChkReXmitMsg(int sRecv, struct fumiedMessage *pMsgRecv,
		 int sSend, struct fumiedMessage *pMsgSend,
		 struct sockaddr_in *addr, u_long waitMsgType,
		 int nretry)
{
  long tm;

  tm = getNextTimeout(INIT_TIMEOUT);
  while ( nretry-- )
    {
      if ( rcvAndCheckFumiedMsg(sRecv, pMsgRecv, waitMsgType, tm) )
	{
	  return 1;
	}
      if ( sendto(sSend, (char *)pMsgSend, sizeof(*pMsgSend), 0,
		  (struct sockaddr *)addr, sizeof(*addr)) < 0 )
	{
	  error("sendto (waitChkReXmitMsg)");
	}
      tm = getNextTimeout(NEXT_TIMEOUT);
    }
  return 0;
}

void
setWaitMsgType(int type, u_int *ptype)
{
  *ptype |= (1 << (type-1));
}

int
if_register_socket(struct interface_info *info)
{
  struct sockaddr_in name;
  int sock;
  int flag;

  /* Set up the address we're going to bind to. */
  name.sin_family = AF_INET;
  name.sin_port = local_port;
  name.sin_addr.s_addr = INADDR_ANY;
  memset (name.sin_zero, 0, sizeof (name.sin_zero));

  /* Make a socket... */
  if ((sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    error ("Can't create fumied socket: %m");

  /* Set the REUSEADDR option so that we don't fail to start if
     we're being restarted. */
  flag = 1;
  if (setsockopt (sock, SOL_SOCKET, SO_REUSEADDR,
		  (char *)&flag, sizeof flag) < 0)
    error ("Can't set SO_REUSEADDR option on fumied socket: %m");

  /* Set the BROADCAST option so that we can broadcast FUMIED responses. */
  if (setsockopt (sock, SOL_SOCKET, SO_BROADCAST,
		  (char *)&flag, sizeof flag) < 0)
    error ("Can't set SO_BROADCAST option on fumied socket: %m");

  /* Bind the socket to this interface's IP address. */
  if (bind (sock, (struct sockaddr *)&name, sizeof name) < 0)
    error ("Can't bind to fumied address: %m");

  /* Bind this socket to this interface. */
  if (setsockopt (sock, SOL_SOCKET, SO_BINDTODEVICE,
		  (char *)(info -> ifp), sizeof *(info -> ifp)) < 0)
    {
      error("setting SO_BINDTODEVICE");
    }

  return sock;
}

void
if_register_send(struct interface_info *info)
{
  info -> wfdesc = info -> rfdesc;
  note ("Sending on   Socket/%s", info -> name);

}

void
if_register_receive(struct interface_info *info)
{
  /* If we're using the socket API for sending and receiving,
     we don't need to register this interface twice. */
  info -> rfdesc = if_register_socket (info);
  note ("Listening on Socket/%s", info -> name);
}

ssize_t
send_packet(struct interface_info *interface, struct packet *packet,
	    struct fumiedMessage *raw, size_t len, struct in_addr from,
	    struct sockaddr_in *to, struct hardware *hto)
{
  int result;

  int retry = 0;
  do
    {
      result = sendto (interface -> wfdesc, (char *)raw, len, 0,
		       (struct sockaddr *)to, sizeof *to);
    }
  while ((to -> sin_addr.s_addr == htonl (INADDR_BROADCAST)) && result < 0
	&& (errno == EHOSTUNREACH || errno == ECONNREFUSED) && retry++ < 10);

  return result;
}

ssize_t
receive_packet(struct interface_info *interface, u_char *buf,
	       size_t len, struct sockaddr_in *from,
	       struct hardware *hfrom)
{
  int flen = sizeof *from;
  int result;

  int retry = 0;
  do
    {
      result = recvfrom (interface -> rfdesc, buf, len, 0,
			 (struct sockaddr *)from, &flen);
    }
  while (result < 0 && (errno == EHOSTUNREACH ||
			errno == ECONNREFUSED) && retry++ < 10);

  return result;
}
